/* OdLib_Io - Copyright (c) 2015 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.mapeadores.opendocument.transform;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Map;
import javax.xml.transform.ErrorListener;
import javax.xml.transform.Source;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.URIResolver;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import net.mapeadores.opendocument.io.ByteArrayOdSource;
import net.mapeadores.opendocument.io.CharSequenceOdSource;
import net.mapeadores.opendocument.io.OdUtils;
import net.mapeadores.opendocument.io.OdZip;
import net.mapeadores.opendocument.io.OdZipEngine;
import net.mapeadores.util.exceptions.NestedTransformerException;
import net.mapeadores.util.io.IOUtils;
import net.mapeadores.util.io.StreamTransformer;
import net.mapeadores.util.io.StreamTransformerErrorHandler;
import net.mapeadores.util.io.TransformerErrorListener;


/**
 *
 * @author Vincent Calame
 */
public class ODStreamTransformerBuilder {

    private final static String HREF_ATTRIBUTE = "xlink:href=";
    private final short odType;
    private StreamTransformerErrorHandler streamTransformerErrorHandler;
    private ErrorListener errorListener;
    private String contentURI;
    private StreamSource streamSource;
    private URIResolver customURIResolver;
    private byte[] stylesByteArray;
    private byte[] contentAutomaticStylesByteArray;
    private PictureProviderBuilder pictureProviderBuilder;

    public ODStreamTransformerBuilder(String extension) {
        this.odType = OdUtils.getOdType(extension);
    }

    public StreamTransformer toStreamTransformer() {
        URIResolver uriResolver = new InternalURIResolver(stylesByteArray, contentAutomaticStylesByteArray, customURIResolver);
        Templates contentTemplate = initTemplates(streamSource, uriResolver);
        TransformerErrorListener transformerErrorListener = new TransformerErrorListener();
        transformerErrorListener.setCustomErrorListener(errorListener);
        return new ODStreamTransformer(odType, contentTemplate, stylesByteArray, uriResolver, transformerErrorListener, pictureProviderBuilder);
    }

    public void setStyles(InputStream inputStream) throws IOException {
        setStyles(IOUtils.toByteArray(inputStream));
    }

    public void setStyles(byte[] byteArray) {
        this.stylesByteArray = byteArray;
    }

    public void setContentAutomaticStyles(InputStream inputStream) throws IOException {
        setContentAutomaticStyles(IOUtils.toByteArray(inputStream));
    }

    public void setContentAutomaticStyles(byte[] byteArray) {
        this.contentAutomaticStylesByteArray = byteArray;
    }

    public void setContentXsl(byte[] byteArray) {
        streamSource = new StreamSource(new ByteArrayInputStream(byteArray));
        streamSource.setSystemId(contentURI);
    }

    public void setContentXsl(String s) {
        streamSource = new StreamSource(new StringReader(s));
        streamSource.setSystemId(contentURI);
    }

    public void setContentXsl(InputStream inputStream) {
        streamSource = new StreamSource(inputStream);
        streamSource.setSystemId(contentURI);
    }

    public void setURIResolver(URIResolver customUriResolver) {
        this.customURIResolver = customUriResolver;
    }

    public void setErrorListener(ErrorListener errorListener) {
        this.errorListener = errorListener;
    }

    public void setStreamTransformerErrorHandler(StreamTransformerErrorHandler streamTransformerErrorHandler) {
        this.streamTransformerErrorHandler = streamTransformerErrorHandler;
    }

    public void setContentURI(String contentURI) {
        this.contentURI = contentURI;
    }

    public void setPictureProviderBuilder(PictureProviderBuilder pictureProviderBuilder) {
        this.pictureProviderBuilder = pictureProviderBuilder;
    }

    private Templates initTemplates(Source source, URIResolver uriResolver) {
        TransformerErrorListener factoryErrorListener = new TransformerErrorListener();
        factoryErrorListener.setCustomErrorListener(errorListener);
        TransformerFactory tFactory = TransformerFactory.newInstance();
        tFactory.setURIResolver(uriResolver);
        tFactory.setErrorListener(factoryErrorListener);
        Templates contentTemplate;
        try {
            contentTemplate = tFactory.newTemplates(source);
        } catch (TransformerException te) {
            factoryErrorListener.checkFatalTransformerException(te);
            contentTemplate = null;
        }
        if (factoryErrorListener.hasError()) {
            if (streamTransformerErrorHandler != null) {
                streamTransformerErrorHandler.wrongResourceError(contentURI, "XSL error :: " + factoryErrorListener.getErrorMessage());
            }
        }
        return contentTemplate;
    }


    private static class ODStreamTransformer implements StreamTransformer {

        private final short odType;
        private final TransformerErrorListener errorListener;
        private final URIResolver uriResolver;
        private final Templates contentTemplate;
        private final byte[] stylesByteArray;
        private final PictureProviderBuilder pictureProviderBuilder;

        private ODStreamTransformer(short odType, Templates contentTemplate, byte[] stylesByteArray, URIResolver uriResolver, TransformerErrorListener errorListener, PictureProviderBuilder pictureProviderBuilder) {
            this.odType = odType;
            this.contentTemplate = contentTemplate;
            this.stylesByteArray = stylesByteArray;
            this.uriResolver = uriResolver;
            this.errorListener = errorListener;
            this.pictureProviderBuilder = pictureProviderBuilder;
        }

        @Override
        public String getMimeType() {
            return OdUtils.getMimeType(odType);
        }

        @Override
        public String getExtension() {
            return OdUtils.getExtension(odType);
        }

        @Override
        public void transform(Source source, OutputStream outputStream, Map<String, String> transformerParameters) throws TransformerException {
            OdZip odZip = OdZip.newInstance(odType);
            if (stylesByteArray != null) {
                odZip.setStylesOdSource(new ByteArrayOdSource(stylesByteArray));
            }
            Transformer transformer = contentTemplate.newTransformer();
            transformer.setURIResolver(uriResolver);
            transformer.setErrorListener(errorListener);
            for (Map.Entry<String, String> entry : transformerParameters.entrySet()) {
                transformer.setParameter(entry.getKey(), entry.getValue());
            }
            StringWriter writer = new StringWriter();
            transformer.transform(source, new StreamResult(writer));
            StringBuffer buf = writer.getBuffer();
            if (pictureProviderBuilder != null) {
                check(buf);
                odZip.setPictureProvider(pictureProviderBuilder.toPictureProvider());
            }
            odZip.setContentOdSource(new CharSequenceOdSource(buf, "UTF-8"));
            try {
                OdZipEngine.run(outputStream, odZip);
            } catch (IOException ioe) {
                throw new TransformerException(ioe);
            } catch (NestedTransformerException nte) {
                throw nte.getTransformerException();
            }
        }

        private void check(StringBuffer buf) {
            int idx = buf.indexOf(HREF_ATTRIBUTE);
            int idx_debut;
            int idx_fin;
            while (idx != -1) {
                idx_debut = idx + HREF_ATTRIBUTE.length() + 1;
                String quoteChar = buf.substring(idx_debut - 1, idx_debut);
                idx_fin = buf.indexOf(quoteChar, idx_debut);
                String path = buf.substring(idx_debut, idx_fin);
                String checkedHref = pictureProviderBuilder.checkHref(path);
                if (checkedHref != null) {
                    buf.delete(idx_debut, idx_fin);
                    buf.insert(idx_debut, checkedHref);
                    idx_fin = idx_debut + checkedHref.length();
                }
                idx = buf.indexOf(HREF_ATTRIBUTE, idx_fin);
            }
        }


    }


    private class InternalURIResolver implements URIResolver {

        private final URIResolver customURIResolver;
        private final byte[] stylesByteArray;
        private final byte[] contentAutomaticStylesByteArray;

        private InternalURIResolver(byte[] stylesByteArray, byte[] contentAutomaticStylesByteArray, URIResolver customURIResolver) {
            this.stylesByteArray = stylesByteArray;
            this.contentAutomaticStylesByteArray = contentAutomaticStylesByteArray;
            this.customURIResolver = customURIResolver;
        }

        @Override
        public Source resolve(String href, String base) throws TransformerException {
            if (href.equals("_content_automaticstyles.xml")) {
                if (contentAutomaticStylesByteArray != null) {
                    return new StreamSource(new ByteArrayInputStream(contentAutomaticStylesByteArray));
                }
            } else if (href.equals("styles.xml")) {
                if (stylesByteArray != null) {
                    return new StreamSource(new ByteArrayInputStream(stylesByteArray));
                }
            }
            if (customURIResolver != null) {
                return customURIResolver.resolve(href, base);
            } else {
                return null;
            }
        }

    }


}
