/* ScrutariLib_SearchEngine - Copyright (c) 2015-2016 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.scrutari.searchengine.tools.operands.search;

import net.mapeadores.util.localisation.LocalisationUtils;
import net.mapeadores.util.logging.MessageHandler;
import net.mapeadores.util.logicaloperation.LogicalOperationParser;
import net.mapeadores.util.logicaloperation.LogicalOperationUtils;
import net.mapeadores.util.logicaloperation.Operand;
import net.mapeadores.util.logicaloperation.OperandException;
import net.mapeadores.util.logicaloperation.SimpleOperand;
import net.mapeadores.util.logicaloperation.SubOperand;
import net.mapeadores.util.text.StringUtils;
import net.mapeadores.util.text.TextConstants;
import net.mapeadores.util.text.search.SearchToken;
import net.mapeadores.util.text.search.SearchUtils;
import net.mapeadores.util.text.search.SimpleSearchToken;
import net.scrutari.db.api.AttributeKeyCoder;
import net.scrutari.searchengine.api.operands.OperandMessageKeys;
import net.scrutari.searchengine.api.operands.search.AlineaFilter;
import net.scrutari.searchengine.api.operands.search.SearchOperation;
import net.scrutari.searchengine.api.operands.search.SearchTokenOperand;


/**
 *
 * @author Vincent Calame
 */
public final class LogicalSearchOperationEngine {

    private final MessageHandler messageHandler;
    private final AttributeKeyCoder attributeKeyCoder;

    private LogicalSearchOperationEngine(MessageHandler messageHandler, AttributeKeyCoder attributeKeyCoder) {
        this.messageHandler = messageHandler;
        this.attributeKeyCoder = attributeKeyCoder;
    }

    public static SearchOperation parse(String q, MessageHandler messageHandler, AttributeKeyCoder attributeKeyCoder) {
        if (q.equals("*")) {
            return SearchOperationBuilder.build(SearchOperationUtils.ALL_OPERAND, 0);
        }
        Operand operand;
        try {
            operand = LogicalOperationParser.parse(q);
        } catch (OperandException oe) {
            addError(messageHandler, oe.getErrorKey(), String.valueOf(oe.getCharIndex()));
            return null;
        }
        LogicalSearchOperationEngine engine = new LogicalSearchOperationEngine(messageHandler, attributeKeyCoder);
        return engine.convert(operand);
    }

    private SearchOperation convert(Operand operand) {
        SearchOperandCheck operandCheck;
        if (operand instanceof SimpleOperand) {
            SearchTokenOperand searchTokenOperand = toSearchTokenOperand(1, (SimpleOperand) operand);
            if (searchTokenOperand == null) {
                return null;
            }
            operandCheck = new SearchOperandCheck(1, searchTokenOperand);
        } else {
            operandCheck = convertSubOperand(1, (SubOperand) operand);
        }
        if (operandCheck == null) {
            return null;
        }
        int operandMaxNumber = operandCheck.getOperandMaxNumber();
        Operand newRootOperand = operandCheck.getOperand();
        return SearchOperationBuilder.build(newRootOperand, operandMaxNumber);
    }

    private SearchOperandCheck convertSubOperand(int currentOperandNumber, SubOperand subOperand) {
        int count = subOperand.getOperandCount();
        Operand[] result = new Operand[count];
        for (int i = 0; i < count; i++) {
            Operand operand = subOperand.getOperand(i);
            if (operand instanceof SimpleOperand) {
                SearchTokenOperand searchTokenOperand = toSearchTokenOperand(currentOperandNumber, (SimpleOperand) operand);
                if (searchTokenOperand == null) {
                    return null;
                }
                result[i] = searchTokenOperand;
                currentOperandNumber++;
            } else {
                SearchOperandCheck check = convertSubOperand(currentOperandNumber, (SubOperand) operand);
                if (check == null) {
                    return null;
                }
                result[i] = check.getOperand();
                currentOperandNumber = check.getOperandMaxNumber() + 1;
            }
        }
        SubOperand resultOperand = LogicalOperationUtils.toSubOperand(result, subOperand.getOperator());
        return new SearchOperandCheck(currentOperandNumber - 1, resultOperand);
    }

    private SearchTokenOperand toSearchTokenOperand(int operandNumber, SimpleOperand simpleOperand) {
        boolean acceptMode = simpleOperand.isAcceptMode();
        String scope = simpleOperand.getScope();
        AlineaFilter alineaFilter = null;
        if (!scope.isEmpty()) {
            alineaFilter = SearchOperationUtils.parse(scope, messageHandler, attributeKeyCoder);
        }
        if (alineaFilter == null) {
            alineaFilter = AlineaFilters.ALL_FILTER;
        }
        SearchToken searchToken = parseSearchToken(simpleOperand.getBody(), simpleOperand.getOperandString());
        if (searchToken == null) {
            return null;
        }
        return SearchOperationUtils.toSearchTokenOperand(operandNumber, acceptMode, searchToken, alineaFilter);
    }

    private SearchToken parseSearchToken(String tokenString, String operandString) {
        int length = tokenString.length();
        if (length == 0) {
            addError(OperandMessageKeys.EMPTY_SEARCHTOKEN, operandString);
            return null;
        }
        boolean onQuote = false;
        if (tokenString.charAt(0) == '"') {
            if (tokenString.charAt(length - 1) != '"') {
                addError(OperandMessageKeys.WRONG_UNCLOSED_QUOTE, operandString);
                return null;
            }
            tokenString = tokenString.substring(1, length - 1).trim();
            length = tokenString.length();
            onQuote = true;
            if (length == 0) {
                addError(OperandMessageKeys.EMPTY_SEARCHTOKEN, operandString);
                return null;
            }
        }
        String[] tokenArray = StringUtils.getTokens(tokenString, ' ', StringUtils.EMPTY_EXCLUDE);
        if (onQuote) {
            int tokenLength = tokenArray.length;
            SimpleSearchToken[] array = new SimpleSearchToken[tokenLength];
            for (int i = 0; i < tokenLength; i++) {
                SimpleSearchToken simpleSearchToken = parseSimpleSearchToken(tokenArray[i], operandString);
                if (simpleSearchToken == null) {
                    return null;
                }
                array[i] = simpleSearchToken;
            }
            return SearchUtils.toMultiSearchToken(array);
        } else {
            if (tokenArray.length > 1) {
                addError(OperandMessageKeys.WRONG_UNEXPECTED_SPACE, operandString);
                return null;
            }
            return parseSimpleSearchToken(tokenArray[0], operandString);
        }
    }

    private SimpleSearchToken parseSimpleSearchToken(String token, String operandString) {
        int length = token.length();
        int startIndex = 0;
        int currentType = TextConstants.MATCHES;
        if (token.charAt(0) == '*') {
            startIndex = 1;
            currentType = TextConstants.ENDSWITH;
        }
        if (token.charAt(length - 1) == '*') {
            length = length - 1;
            currentType = currentType | TextConstants.STARTSWITH;
        }
        if (startIndex >= length) {
            addError(OperandMessageKeys.WRONG_MISSING_WORD, operandString);
            return null;
        }
        StringBuilder stringBuilder = new StringBuilder();
        for (int i = startIndex; i < length; i++) {
            char carac = token.charAt(i);
            if (!Character.isLetterOrDigit(carac)) {
                addError(OperandMessageKeys.WRONG_NOT_LETTERORDIGIT_CHARACTER, operandString);
                return null;
            }
            stringBuilder.append(carac);
        }
        return SearchUtils.toSimpleSearchToken(stringBuilder.toString(), currentType);
    }

    private void addError(String key, Object value) {
        addError(messageHandler, OperandMessageKeys.ERROR_CATEGORY, LocalisationUtils.toMessage(key, value));
    }

    private static void addError(MessageHandler messageHandler, String key, Object value) {
        messageHandler.addMessage(OperandMessageKeys.ERROR_CATEGORY, LocalisationUtils.toMessage(key, value));
    }

}
