/* SctServer - Copyright (c) 2013-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.sctserver.tools.collect;

import fr.exemole.sctserver.api.collect.CollectResult;
import fr.exemole.sctserver.api.collect.ScrutariInfoContent;
import fr.exemole.sctserver.api.collect.ScrutariSource;
import fr.exemole.sctserver.api.collect.ScrutariSourceDef;
import fr.exemole.sctserver.api.collect.StateErrorException;
import fr.exemole.sctserver.api.storage.EngineStorage;
import fr.exemole.sctserver.api.storage.SctStorageException;
import fr.exemole.sctserver.tools.ErrorUtils;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import net.mapeadores.util.exceptions.NestedLibraryException;
import net.mapeadores.util.exceptions.SwitchException;
import net.mapeadores.util.logging.MessageHandler;
import net.mapeadores.util.logging.MultiMessageHandler;
import net.mapeadores.util.primitives.FuzzyDate;
import net.mapeadores.util.text.StringUtils;
import net.mapeadores.util.xml.XMLUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;


/**
 *
 * @author Vincent Calame
 */
public final class CollectEngine {

    private final ScrutariSource scrutariSource;
    private final String scrutariSourceName;
    private final EngineStorage engineStorage;
    private final MultiMessageHandler messageHandler;
    private final ScrutariSourceDef scrutariSourceDef;
    private final boolean forceDataDownload;
    private final List<EngineStorage.MatchedData> matchedDataList = new ArrayList<EngineStorage.MatchedData>();
    private FuzzyDate lastUpdateDate = null;


    static {
        CollectUtils.initSslFactory();
    }

    private CollectEngine(ScrutariSource scrutariSource, EngineStorage engineStorage, MultiMessageHandler messageHandler, boolean forceDataDownload) {
        this.engineStorage = engineStorage;
        this.scrutariSource = scrutariSource;
        this.scrutariSourceName = scrutariSource.getScrutariSourceDef().getName();
        this.messageHandler = messageHandler;
        this.forceDataDownload = forceDataDownload;
        this.scrutariSourceDef = scrutariSource.getScrutariSourceDef();
    }


    private CollectResult run() {
        short state = CollectResult.OK_STATE;
        messageHandler.setCurrentURI(scrutariSourceDef.getName());
        boolean error = false;
        for (ScrutariSourceDef.UrlEntry urlEntry : scrutariSourceDef.getUrlEntryList()) {
            short type = urlEntry.getType();
            try {
                switch (type) {
                    case ScrutariSourceDef.INFO_TYPE:
                        fromInfoType(urlEntry);
                        break;
                    case ScrutariSourceDef.SCRUTARIDATA_TYPE:
                        fromDataType(urlEntry);
                        break;
                    default:
                        throw new SwitchException("Unknown type: " + type);
                }
            } catch (StateErrorException stateErrorException) {
                state = stateErrorException.getErrorState();
                lastUpdateDate = null;
                error = true;
            }
            if (error) {
                break;
            }
        }
        if (!error) {
            engineStorage.saveScrutariData(scrutariSource.getScrutariSourceDef().getName(), matchedDataList);
        }
        for (EngineStorage.MatchedData data : matchedDataList) {
            ((InternalMatchedData) data).delete();
        }
        return CollectUtils.toCollectResult(state, lastUpdateDate);
    }

    public static CollectResult collectData(ScrutariSource scrutariSource, EngineStorage engineStorage, MultiMessageHandler messageHandler, boolean forceDataDownload) {
        messageHandler.setCurrentURI(scrutariSource.getScrutariSourceDef().getName());
        CollectEngine engine = new CollectEngine(scrutariSource, engineStorage, messageHandler, forceDataDownload);
        return engine.run();
    }


    private void fromDataType(ScrutariSourceDef.UrlEntry urlEntry) throws StateErrorException {
        URL dataUrl = urlEntry.getUrl();
        int frequency = urlEntry.getFrequency();
        FuzzyDate currentDate = FuzzyDate.current();
        if ((!forceDataDownload) && (FuzzyDate.compare(scrutariSource.getLastUpdate(), currentDate.roll(-frequency))) > 0) {
            lastUpdateDate = FuzzyDate.max(lastUpdateDate, scrutariSource.getLastUpdate());
            if (engineStorage.hasData(scrutariSourceName, dataUrl)) {
                addNoChange(dataUrl);
                return;
            }
        }
        File tempFile = createTempFile(scrutariSourceName);
        try (OutputStream os = new FileOutputStream(tempFile)) {
            downloadData(os, dataUrl, messageHandler);
        } catch (IOException ioe) {
            tempFile.delete();
            throw new SctStorageException(ioe);
        }
        addChange(dataUrl, tempFile);
        lastUpdateDate = currentDate;
    }

    private void fromInfoType(ScrutariSourceDef.UrlEntry urlEntry) throws StateErrorException {
        URL infoUrl = urlEntry.getUrl();
        ScrutariInfoContent scrutariInfoContent = getInfoContent(infoUrl, messageHandler);
        FuzzyDate infoLastUpdate = scrutariInfoContent.getLastUpdate();
        if (!forceDataDownload) {
            if (FuzzyDate.compare(infoLastUpdate, scrutariSource.getLastUpdate()) < 0) {
                lastUpdateDate = FuzzyDate.max(lastUpdateDate, scrutariSource.getLastUpdate());
                if (engineStorage.hasData(scrutariSourceName, infoUrl)) {
                    addNoChange(infoUrl);
                    return;
                }
            }
        }
        File tempFile = createTempFile(scrutariSourceName);
        try (OutputStream os = new FileOutputStream(tempFile)) {
            for (URL dataUrl : scrutariInfoContent.getDataUrlList()) {
                downloadData(os, dataUrl, messageHandler);
            }
        } catch (IOException ioe) {
            throw new SctStorageException(ioe);
        }
        addChange(infoUrl, tempFile);
        lastUpdateDate = FuzzyDate.current();
    }

    private void addChange(URL url, File tempFile) {
        matchedDataList.add(new InternalMatchedData(url, tempFile));
    }

    private void addNoChange(URL url) {
        matchedDataList.add(new InternalMatchedData(url, null));
    }

    private static File createTempFile(String name) {
        try {
            File tempFile = File.createTempFile(name, ".tmp");
            tempFile.deleteOnExit();
            return tempFile;
        } catch (IOException ioe) {
            throw new SctStorageException(ioe);
        }
    }

    private static void downloadData(OutputStream outputStream, URL dataUrl, MultiMessageHandler messageHandler) throws IOException, StateErrorException {
        URLConnection con = null;
        try {
            con = dataUrl.openConnection();
            if (!CollectUtils.testConnection(con, messageHandler)) {
                throw new StateErrorException(CollectResult.DATA_BROKENURL_STATE);
            }
        } catch (IOException ioe) {
            severeDownload(messageHandler, dataUrl, ioe.getMessage());
            throw new StateErrorException(CollectResult.DATA_NOCONNECTION_STATE);
        }
        try (InputStream is = con.getInputStream()) {
            byte[] buffer = new byte[1024];
            int size;
            while ((size = is.read(buffer, 0, 1024)) > 0) {
                outputStream.write(buffer, 0, size);
            }
        } catch (IOException ioe) {
            severeDownload(messageHandler, dataUrl, ioe.getMessage());
            throw new StateErrorException(CollectResult.DATA_DOWNLOADERROR_STATE);
        }

    }

    private static ScrutariInfoContent getInfoContent(URL url, MultiMessageHandler messageHandler) throws StateErrorException {
        URLConnection con = null;
        try {
            con = (HttpURLConnection) url.openConnection();
            if (!CollectUtils.testConnection(con, messageHandler)) {
                throw new StateErrorException(CollectResult.INFO_BROKENURL_STATE);
            }
        } catch (IOException ioe) {
            severeDownload(messageHandler, url, ioe.getMessage());
            throw new StateErrorException(CollectResult.INFO_NOCONNECTION_STATE);
        }
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder db = null;
        try {
            db = dbf.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new NestedLibraryException(e);
        }
        Document doc = null;
        try {
            doc = db.parse(con.getInputStream());
        } catch (IOException ioe) {
            severeDownload(messageHandler, url, ioe.getMessage());
            throw new StateErrorException(CollectResult.INFO_DOWNLOADERROR_STATE);
        } catch (SAXException saxe) {
            severeDownload(messageHandler, url, saxe.getMessage());
            throw new StateErrorException(CollectResult.INFO_NOTXML_STATE);
        }
        String infoUrlString = url.toString();
        int idx = infoUrlString.lastIndexOf('/');
        String urlRoot = infoUrlString.substring(0, idx + 1);
        return readScrutariInfoDOM(doc.getDocumentElement(), urlRoot, messageHandler);
    }

    private static ScrutariInfoContent readScrutariInfoDOM(Element element_xml, String urlRoot, MultiMessageHandler messageHandler) throws StateErrorException {
        String lastDateString = null;
        List<String> dataUrlStringList = new ArrayList<String>();
        NodeList nodeList = element_xml.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node node = nodeList.item(i);
            if (node.getNodeType() == Node.ELEMENT_NODE) {
                Element el = (Element) node;
                String tagname = el.getTagName();
                if (tagname.equals("date")) {
                    lastDateString = XMLUtils.getData(el);
                } else if ((tagname.equals("path")) || (tagname.equals("scrutaridata-url"))) {
                    String pathString = XMLUtils.getData(el);
                    if (pathString.length() > 0) {
                        dataUrlStringList.add(pathString);
                    }
                }
            }
        }
        boolean wrongXml = false;
        if (lastDateString == null) {
            ErrorUtils.severe(messageHandler, "_ error.unsupported.xml.missingtag", element_xml.getTagName(), "date");
            wrongXml = true;
        }
        if (dataUrlStringList.isEmpty()) {
            ErrorUtils.severe(messageHandler, "_ error.unsupported.xml.missingtag", element_xml.getTagName(), "scrutaridata-url");
            wrongXml = true;
        }
        if (wrongXml) {
            throw new StateErrorException(CollectResult.INFO_WRONGXML_STATE);
        }
        FuzzyDate lastUpdate;
        try {
            lastUpdate = FuzzyDate.parse(lastDateString);
        } catch (ParseException pe) {
            ErrorUtils.severe(messageHandler, "_ eerror.wrong.xml.tagvalue", "date", lastDateString);
            throw new StateErrorException(CollectResult.INFO_WRONGXML_STATE);
        }
        List<URL> urlList = new ArrayList<URL>();
        for (String urlString : dataUrlStringList) {
            try {
                URL dataUrl = buildDataUrl(urlString, urlRoot);
                urlList.add(dataUrl);
            } catch (MalformedURLException mue) {
                ErrorUtils.severe(messageHandler, "_ eerror.wrong.xml.tagvalue", "scrutaridata-url", urlString);
                throw new StateErrorException(CollectResult.INFO_WRONGXML_STATE);
            }
        }
        return new InternalScrutariInfoContent(lastUpdate, StringUtils.wrap(urlList.toArray(new URL[urlList.size()])));
    }

    private static URL buildDataUrl(String pathString, String urlRoot) throws MalformedURLException {
        try {
            URL url = new URL(pathString);
            return url;
        } catch (MalformedURLException mue) {
            URL url = new URL(urlRoot + pathString);
            return url;
        }
    }

    private static void severeDownload(MessageHandler errorHandler, URL url, String message) {
        ErrorUtils.severe(errorHandler, "_ error.exception.download", url.toString(), message);
    }


    private static class InternalScrutariInfoContent implements ScrutariInfoContent {

        private final FuzzyDate lastUpdate;
        private final List<URL> urlList;

        private InternalScrutariInfoContent(FuzzyDate lastUpdate, List<URL> urlList) {
            this.lastUpdate = lastUpdate;
            this.urlList = urlList;
        }

        @Override
        public FuzzyDate getLastUpdate() {
            return lastUpdate;
        }

        @Override
        public List<URL> getDataUrlList() {
            return urlList;
        }

    }


    private static class InternalMatchedData implements EngineStorage.MatchedData {

        private final File tempFile;
        private final URL url;

        private InternalMatchedData(URL url, File tempFile) {
            this.tempFile = tempFile;
            this.url = url;
        }

        @Override
        public InputStream getInputStream() {
            if (tempFile == null) {
                return null;
            }
            try {
                return new FileInputStream(tempFile);
            } catch (IOException ioe) {
                throw new SctStorageException(ioe);
            }
        }

        @Override
        public URL getMatchingURL() {
            return url;
        }

        @Override
        public boolean isNotChanged() {
            return (tempFile == null);
        }

        private void delete() {
            if (tempFile != null) {
                tempFile.delete();
            }
        }

    }

}
