/* SctServer_Directory - Copyright (c) 2014 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.sctserver.directory.cache;

import fr.exemole.sctserver.api.storage.SctStorageException;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import net.mapeadores.util.localisation.Lang;
import net.mapeadores.util.primitives.io.DataInputPrimitivesReader;
import net.scrutari.db.api.caches.CacheIterator;
import net.scrutari.db.api.caches.LexieCacheReader;
import net.scrutari.db.api.caches.LexieCacheReaderFactory;
import net.scrutari.lexie.FicheLexification;
import net.scrutari.lexie.LexieOccurrences;
import net.scrutari.lexie.LexieUtils;
import net.scrutari.lexie.MotcleLexification;
import net.scrutari.lexie.ScrutariLexieUnit;


/**
 *
 * @author Vincent Calame
 */
public class DirectoryLexieCacheReaderFactory implements LexieCacheReaderFactory {

    private final File occurrencesFile;
    private final File lexificationFile;
    private final File unitFile;
    private final Map<Integer, Long> lexificationPositionMap;
    private final Map<OccurrencesKey, Long> occurrencesPositionMap;
    private final Map<Lang, Long> unitPositionMap;

    public DirectoryLexieCacheReaderFactory(File occurrencesFile, File lexificationFile, File unitFile, Map<Integer, Long> lexificationPositionMap, Map<OccurrencesKey, Long> occurrencesPositionMap, Map<Lang, Long> unitPositionMap) {
        this.occurrencesFile = occurrencesFile;
        this.lexificationFile = lexificationFile;
        this.unitFile = unitFile;
        this.lexificationPositionMap = lexificationPositionMap;
        this.occurrencesPositionMap = occurrencesPositionMap;
        this.unitPositionMap = unitPositionMap;
    }

    @Override
    public LexieCacheReader newReader() {
        return new InternalLexieCacheReader();
    }

    @Override
    public CacheIterator<LexieCacheReaderFactory.LexieUnitInfo> newLexieUnitInfoIterator() {
        return CacheUtils.newLexieUnitInfoIterator(unitFile);
    }

    public static LexieCacheReaderFactory loadFromCache(File cacheDirectory) {
        File occurrencesFile = new File(cacheDirectory, "lexie-occurrences.primitives");
        if (!occurrencesFile.exists()) {
            return null;
        }
        File lexificationFile = new File(cacheDirectory, "lexie-lexification.primitives");
        if (!lexificationFile.exists()) {
            return null;
        }
        File unitFile = new File(cacheDirectory, "lexie-unit.primitives");
        if (!unitFile.exists()) {
            return null;
        }
        Map<Integer, Long> lexificationPositionMap = new HashMap<Integer, Long>();
        Map<OccurrencesKey, Long> occurrencesPositionMap = new HashMap<OccurrencesKey, Long>();
        Map<Lang, Long> unitPositionMap = new HashMap<Lang, Long>();
        try {
            CacheUtils.checkLexificationPosition(lexificationFile, lexificationPositionMap);
            CacheUtils.checkOccurrencePosition(occurrencesFile, occurrencesPositionMap);
            CacheUtils.checkUnitPosition(unitFile, unitPositionMap);
        } catch (IOException ioe) {
            return null;
        }
        return new DirectoryLexieCacheReaderFactory(occurrencesFile, lexificationFile, unitFile, lexificationPositionMap, occurrencesPositionMap, unitPositionMap);
    }


    private class InternalLexieCacheReader implements LexieCacheReader {

        private RandomAccessFile occurrencesRandomAccessFile;
        private DataInputPrimitivesReader occurrencesPrimitivesReader;
        private RandomAccessFile lexificationRandomAccessFile;
        private DataInputPrimitivesReader lexificationPrimitivesReader;
        private RandomAccessFile unitRandomAccessFile;
        private DataInputPrimitivesReader unitPrimitivesReader;

        private InternalLexieCacheReader() {
            try {
                this.unitRandomAccessFile = new RandomAccessFile(unitFile, "r");
            } catch (IOException ioe) {
                throw new SctStorageException(ioe);
            }
            this.unitPrimitivesReader = new DataInputPrimitivesReader(unitRandomAccessFile);
        }

        @Override
        public List<ScrutariLexieUnit> getScrutariLexieUnitList(Lang langInteger) {
            Long position = unitPositionMap.get(langInteger);
            if (position == null) {
                return LexieUtils.EMPTY_SCRUTARILEXIEUNITLIST;
            }
            try {
                if (unitRandomAccessFile == null) {
                    unitRandomAccessFile = new RandomAccessFile(unitFile, "r");
                    unitPrimitivesReader = new DataInputPrimitivesReader(unitRandomAccessFile);
                }
                unitRandomAccessFile.seek(position);
                return CacheUtils.scrutariLexieUnitListFromPrimitives(unitPrimitivesReader);
            } catch (IOException ioe) {
                throw new SctStorageException(ioe);
            }
        }

        @Override
        public List<LexieOccurrences> getLexieOccurrencesList(long lexieId, int subsetCode) {
            Long position = occurrencesPositionMap.get(new OccurrencesKey(lexieId, subsetCode));
            if (position == null) {
                return LexieUtils.EMPTY_LEXIEOCCURRENCESLIST;
            }
            try {
                if (occurrencesRandomAccessFile == null) {
                    occurrencesRandomAccessFile = new RandomAccessFile(occurrencesFile, "r");
                    occurrencesPrimitivesReader = new DataInputPrimitivesReader(occurrencesRandomAccessFile);
                }
                occurrencesRandomAccessFile.seek(position);
                return CacheUtils.lexieOccurrencesListFromPrimitives(occurrencesPrimitivesReader);
            } catch (IOException ioe) {
                throw new SctStorageException(ioe);
            }
        }

        @Override
        public FicheLexification getFicheLexification(int codefiche) {
            try {
                boolean done = testLexification(codefiche);
                if (!done) {
                    return null;
                }
                return FicheLexification.fromPrimitives(lexificationPrimitivesReader);
            } catch (IOException ioe) {
                throw new SctStorageException(ioe);
            }
        }

        @Override
        public MotcleLexification getMotcleLexification(int codemotcle) {
            try {
                boolean done = testLexification(codemotcle);
                if (!done) {
                    return null;
                }
                return MotcleLexification.fromPrimitives(lexificationPrimitivesReader);
            } catch (IOException ioe) {
                throw new SctStorageException(ioe);
            }
        }

        @Override
        public void close() {
            try {
                if (occurrencesRandomAccessFile != null) {
                    occurrencesPrimitivesReader = null;
                    occurrencesRandomAccessFile.close();
                    occurrencesRandomAccessFile = null;
                }
                if (lexificationRandomAccessFile != null) {
                    lexificationPrimitivesReader = null;
                    lexificationRandomAccessFile.close();
                    lexificationRandomAccessFile = null;
                }
                if (unitRandomAccessFile != null) {
                    unitPrimitivesReader = null;
                    unitRandomAccessFile.close();
                    unitRandomAccessFile = null;
                }
            } catch (IOException ioe) {
                throw new SctStorageException(ioe);
            }
        }

        private boolean testLexification(int elementCode) throws IOException {
            Long position = lexificationPositionMap.get(elementCode);
            if (position == null) {
                return false;
            }
            if (lexificationRandomAccessFile == null) {
                lexificationRandomAccessFile = new RandomAccessFile(lexificationFile, "r");
                lexificationPrimitivesReader = new DataInputPrimitivesReader(lexificationRandomAccessFile);
            }
            lexificationRandomAccessFile.seek(position);
            return true;
        }

    }

}
