/* UtilLib - Copyright (c) 2008-2017 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.mapeadores.util.html;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import net.mapeadores.util.localisation.Lang;


/**
 *
 * @author Vincent Calame
 */
public class PageHtmlPrinter extends HtmlPrinter {

    private String htmlpage_title;
    private HtmlAttributes bodyHTMLAttributes;
    private CommandBoxParameters commandBoxParameters;
    private boolean onCommandBox = false;
    private String currentCommandKey = null;
    private final List<String> cssUrlList = new ArrayList<String>();
    private final List<String> jsUrlList = new ArrayList<String>();
    private String iconPng = null;

    public PageHtmlPrinter() {
    }

    public void setCommandBoxParameters(CommandBoxParameters commandBoxParameters) {
        this.commandBoxParameters = commandBoxParameters;
    }

    public CommandBoxParameters getCommandBoxParameters() {
        return commandBoxParameters;
    }

    public void setTitle(String title) {
        this.htmlpage_title = title;
    }

    public String getTitle() {
        return htmlpage_title;
    }

    public void setBodyHTMLAttributes(HtmlAttributes bodyHTMLAttributes) {
        this.bodyHTMLAttributes = bodyHTMLAttributes;
    }

    public void addCssUrl(String cssUrl) {
        cssUrlList.add(cssUrl);
    }

    public void addJsUrl(String jsUrl) {
        jsUrlList.add(jsUrl);
    }

    public void start(Lang lang) {
        this
                .DOCTYPE_html5()
                .HTML(lang);
        this
                .HEAD();
        this
                .TITLE()
                .escape(htmlpage_title)
                ._TITLE();
        this
                .META_htmlcontenttype();
        if (!jsUrlList.isEmpty()) {
            this
                    .META(HA.httpEquiv("Content-Script-Type").content("text/javascript"));
        }
        for (String cssUrl : cssUrlList) {
            this
                    .LINK(HA.type("text/css").href(cssUrl).rel("stylesheet"));
        }
        if (iconPng != null) {
            this
                    .LINK(HA.type("image/png").href(iconPng).rel("icon"));
        }
        if (!jsUrlList.isEmpty()) {
            HtmlAttributes scriptAttributes = HA.type("text/javascript");
            int length = jsUrlList.size();
            for (int i = 0; i < length; i++) {
                scriptAttributes.src(jsUrlList.get(i));
                this
                        .SCRIPT(scriptAttributes)
                        ._SCRIPT();
            }
        }
        this
                ._HEAD();
        this
                .BODY(bodyHTMLAttributes);
    }

    public void startCommandBox(CommandBox commandBox) {
        if (onCommandBox) {
            throw new IllegalStateException("last CommandBox is not ended");
        }
        onCommandBox = true;
        HtmlAttributes formAttributes = HtmlPrinterUtils.newFormAttributes(commandBoxParameters);
        formAttributes.action(commandBox.action());
        if (commandBox.multipart()) {
            formAttributes.enctype(HtmlConstants.MULTIPART_ENCTYPE);
        }
        String target = commandBox.target();
        if (target != null) {
            formAttributes.target(target);
        }
        this
                .FORM_post(formAttributes);
        this
                .INPUT_hidden("cmd", commandBox.name());
        String page = commandBox.page();
        if (page != null) {
            this
                    .INPUT_hidden("page", page);
        }
        String errorPage = commandBox.errorPage();
        if (errorPage != null) {
            this
                    .INPUT_hidden("page-err", errorPage);
        }
        Map<String, String> hiddenAttributesMap = commandBox.hiddenMap();
        if (hiddenAttributesMap != null) {
            this
                    .INPUT_hidden(hiddenAttributesMap);
        }
        HtmlAttributes fieldsetAttributes = HtmlPrinterUtils.newFieldsetAttributes(commandBoxParameters);
        String commandFamily = commandBox.family();
        if (commandFamily != null) {
            fieldsetAttributes.attr("data-command-family", commandFamily);
        } else {
            fieldsetAttributes.attr("data-command-family", null);
        }
        this
                .FIELDSET(fieldsetAttributes);
        String commandKey = commandBox.lockey();
        if (commandKey != null) {
            this
                    .LEGEND(HtmlPrinterUtils.newLegendAttributes(commandBoxParameters))
                    .localize(commandKey)
                    ._LEGEND();
            currentCommandKey = commandKey;
        }
    }

    public void endCommandBox(String submitMessageKey) {
        if (!onCommandBox) {
            throw new IllegalStateException("no CommandBox started yet");
        }
        if (submitMessageKey != null) {
            this
                    .DIV(HtmlPrinterUtils.newSubmitDivAttributes(commandBoxParameters))
                    .INPUT_submit(getLocalization(submitMessageKey))
                    ._DIV();
        }
        if (currentCommandKey != null) {
            String keyString = currentCommandKey;
            if (keyString.startsWith("_ ")) {
                keyString = keyString.substring(2);
            }
            this
                    .DIV(HtmlPrinterUtils.newCommandKeyDivAttributes(commandBoxParameters))
                    .escape(keyString)
                    ._DIV();
        }
        this
                ._FIELDSET();
        this
                ._FORM();
        currentCommandKey = null;
        onCommandBox = false;
    }

    public void end() {
        if (onCommandBox) {
            throw new IllegalStateException("last CommandBox is not ended");
        }
        this
                ._BODY();
        this
                ._HTML();
    }

    public void setIconPng(String iconPng) {
        this.iconPng = iconPng;
    }

}
